from .interface import BaseInteriorPointInterface, InteriorPointInterface
from .schur_complement.sc_ip_interface import DynamicSchurComplementInteriorPointInterface
from .schur_complement.mpi_sc_ip_interface import MPIDynamicSchurComplementInteriorPointInterface
